/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  GMime
 *  Copyright (C) 2000-2010 Jeffrey Stedfast
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1
 *  of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
 *  02110-1301, USA.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>

#include "url-scanner.h"
#include "gmime-filter-html.h"

#define d(x)


/**
 * SECTION: gmime-filter-html
 * @title: GMimeFilterHTML
 * @short_description: Convert plain text into HTML
 * @see_also: #GMimeFilter
 *
 * A #GMimeFilter used for converting plain text into HTML.
 **/


#define CONVERT_WEB_URLS  GMIME_FILTER_HTML_CONVERT_URLS
#define CONVERT_ADDRSPEC  GMIME_FILTER_HTML_CONVERT_ADDRESSES

static struct {
	unsigned int mask;
	urlpattern_t pattern;
} patterns[] = {
	{ CONVERT_WEB_URLS, { "file://",   "",        url_file_start,     url_file_end     } },
	{ CONVERT_WEB_URLS, { "ftp://",    "",        url_web_start,      url_web_end      } },
	{ CONVERT_WEB_URLS, { "http://",   "",        url_web_start,      url_web_end      } },
	{ CONVERT_WEB_URLS, { "https://",  "",        url_web_start,      url_web_end      } },
	{ CONVERT_WEB_URLS, { "news://",   "",        url_web_start,      url_web_end      } },
	{ CONVERT_WEB_URLS, { "nntp://",   "",        url_web_start,      url_web_end      } },
	{ CONVERT_WEB_URLS, { "telnet://", "",        url_web_start,      url_web_end      } },
	{ CONVERT_WEB_URLS, { "www.",      "http://", url_web_start,      url_web_end      } },
	{ CONVERT_WEB_URLS, { "ftp.",      "ftp://",  url_web_start,      url_web_end      } },
	{ CONVERT_ADDRSPEC, { "@",         "mailto:", url_addrspec_start, url_addrspec_end } },
};

static const gchar preamble[] = "<html>\n<head>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"><meta name=\"Generator\" content=\"GMime Library\">\n</head>\n<body>\n";
static const gchar postamble[] = "\n</body>\n</html>\n";

static const gchar br[] = "<br>";
static const gchar div_start[] = "<div>";
static const gchar div_end[] = "</div>";
static const gchar p_start[] = "<p>";
static const gchar p_end[] = "</p>";


#define PREAMBLE_SIZE (G_N_ELEMENTS(preamble))
#define POSTAMBLE_SIZE (G_N_ELEMENTS(postamble))
#define MAX_NL_SIZE (G_N_ELEMENTS(div_end))

#define NUM_URL_PATTERNS (sizeof (patterns) / sizeof (patterns[0]))

static void g_mime_filter_html_class_init (GMimeFilterHTMLClass *klass);
static void g_mime_filter_html_init (GMimeFilterHTML *filter, GMimeFilterHTMLClass *klass);
static void g_mime_filter_html_finalize (GObject *object);

static GMimeFilter *filter_copy (GMimeFilter *filter);
static void filter_filter (GMimeFilter *filter, char *in, size_t len, size_t prespace,
			   char **out, size_t *outlen, size_t *outprespace);
static void filter_complete (GMimeFilter *filter, char *in, size_t len, size_t prespace,
			     char **out, size_t *outlen, size_t *outprespace);
static void filter_reset (GMimeFilter *filter);

static const gchar *get_nl(gint32 nl_style, gboolean start);

static GMimeFilterClass *parent_class = NULL;


GType
g_mime_filter_html_get_type (void)
{
	static GType type = 0;
	
	if (!type) {
		static const GTypeInfo info = {
			sizeof (GMimeFilterHTMLClass),
			NULL, /* base_class_init */
			NULL, /* base_class_finalize */
			(GClassInitFunc) g_mime_filter_html_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (GMimeFilterHTML),
			0,    /* n_preallocs */
			(GInstanceInitFunc) g_mime_filter_html_init,
		};
		
		type = g_type_register_static (GMIME_TYPE_FILTER, "GMimeFilterHTML", &info, 0);
	}
	
	return type;
}


static void
g_mime_filter_html_class_init (GMimeFilterHTMLClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GMimeFilterClass *filter_class = GMIME_FILTER_CLASS (klass);
	
	parent_class = g_type_class_ref (GMIME_TYPE_FILTER);
	
	object_class->finalize = g_mime_filter_html_finalize;
	
	filter_class->copy = filter_copy;
	filter_class->filter = filter_filter;
	filter_class->complete = filter_complete;
	filter_class->reset = filter_reset;
}

static void
g_mime_filter_html_init (GMimeFilterHTML *filter, GMimeFilterHTMLClass *klass)
{
	filter->scanner = url_scanner_new ();
	
	filter->flags = 0;
	filter->colour = 0;
	filter->column = 0;
	filter->pre_open = FALSE;
    filter->first_call = TRUE;
}

static void
g_mime_filter_html_finalize (GObject *object)
{
	GMimeFilterHTML *html = (GMimeFilterHTML *) object;
	
	url_scanner_free (html->scanner);
	
	G_OBJECT_CLASS (parent_class)->finalize (object);
}


static GMimeFilter *
filter_copy (GMimeFilter *filter)
{
	GMimeFilterHTML *html = (GMimeFilterHTML *) filter;
	
	return g_mime_filter_html_new (html->flags, html->colour, html->nl_style);
}

static char *
check_size (GMimeFilter *filter, char *outptr, char **outend, size_t len)
{
	size_t outleft = (size_t) (*outend - outptr);
	size_t offset;
	
	if (outleft >= len)
		return outptr;
	
	offset = outptr - filter->outbuf;
	
	g_mime_filter_set_size (filter, filter->outsize + len, TRUE);
	
	*outend = filter->outbuf + filter->outsize;
	
	return filter->outbuf + offset;
}

static int
citation_depth (const char *in, const char *inend)
{
    register const char *inptr = in;
    int depth = 1;

    if (inptr >= inend) {
        return 0;
    }

    if (*inptr++ != '>')
        return 0;

    /* check that it isn't an escaped From line */
    if (inptr + 4 <= inend && !strncmp (inptr, "From", 4))
        return 0;

    while (inptr < inend && *inptr != '\n') {
        if (*inptr == ' ')
            inptr++;

        if (inptr >= inend || *inptr++ != '>')
            break;

        depth++;
    }

    return depth;
}

static inline gunichar
html_utf8_getc (const unsigned char **in, const unsigned char *inend)
{
	register const unsigned char *inptr = *in;
	register unsigned char c, r;
	register gunichar u, m;
	
	if (inptr == inend)
		return 0;
	
	while (inptr < inend) {
		r = *inptr++;
	loop:
		if (r < 0x80) {
			*in = inptr;
			return r;
		} else if (r < 0xf8) { /* valid start char? */
			u = r;
			m = 0x7f80;	/* used to mask out the length bits */
			do {
				if (inptr >= inend)
					return 0xffff;
				
				c = *inptr++;
				if ((c & 0xc0) != 0x80) {
					r = c;
					goto loop;
				}
				
				u = (u << 6) | (c & 0x3f);
				r <<= 1;
				m <<= 5;
			} while (r & 0x40);
			
			*in = inptr;
			
			u &= ~m;
			
			return u;
		}
	}
	
	return 0xffff;
}

static char *
writeln (GMimeFilter *filter, const char *in, const char *end, char *outptr, char **outend)
{
	GMimeFilterHTML *html = (GMimeFilterHTML *) filter;
	const unsigned char *instart = (const unsigned char *) in;
	const unsigned char *inend = (const unsigned char *) end;
	const unsigned char *inptr = instart;
    const unsigned char *char_start;
	
	while (inptr < inend) {
		gunichar u;
		
		outptr = check_size (filter, outptr, outend, 16);
		
        char_start = inptr;
		u = html_utf8_getc (&inptr, inend);
		switch (u) {
		case 0xffff:
            // Petr Silar: Warning commented because this may occur on the end of the chunk.
			//g_warning ("Invalid UTF-8 sequence encountered");
			return outptr;
			break;
		case '<':
			outptr = g_stpcpy (outptr, "&lt;");
			html->column++;
			break;
		case '>':
			outptr = g_stpcpy (outptr, "&gt;");
			html->column++;
			break;
		case '&':
			outptr = g_stpcpy (outptr, "&amp;");
			html->column++;
			break;
		case '"':
			outptr = g_stpcpy (outptr, "&quot;");
			html->column++;
			break;
        case 0xa0:
        case 0xc2a0:
            outptr = g_stpcpy(outptr, "&nbsp;");
            html->column++;
            break;
        case '\t':
			if (html->flags & (GMIME_FILTER_HTML_CONVERT_SPACES)) {
				do {
					outptr = check_size (filter, outptr, outend, 7);
					outptr = g_stpcpy (outptr, "&nbsp;");
					html->column++;
				} while (html->column % 8);
				break;
			}
			/* otherwise, FALL THROUGH */
		case ' ':
			if (html->flags & GMIME_FILTER_HTML_CONVERT_SPACES) {
				if (inptr == (instart + 1) || (inptr < inend && (*inptr == ' ' || *inptr == '\t'))) {
					outptr = g_stpcpy (outptr, "&nbsp;");
					html->column++;
					break;
				}
			}
			/* otherwise, FALL THROUGH */
		default:
			if (u >= 0x20 && u < 0x80) {
				*outptr++ = (char) (u & 0xff);
			} else {
				if (html->flags & GMIME_FILTER_HTML_ESCAPE_8BIT)
					*outptr++ = '?';
                else if (html->flags & GMIME_FILTER_HTML_UTF8) {
                    memcpy(outptr, char_start, inptr - char_start);
                    outptr += inptr - char_start;
                    *outptr = 0;
                }
                else
                    outptr += sprintf (outptr, "&#%u;", u);
			}
			html->column++;
			break;
		}
	}
	
	return outptr;
}

static void
html_convert (GMimeFilter *filter, char *in, size_t inlen, size_t prespace,
	      char **out, size_t *outlen, size_t *outprespace, gboolean flush)
{
	GMimeFilterHTML *html = (GMimeFilterHTML *) filter;
	register char *inptr, *outptr;
	char *start, *outend;
	char *inend, *realinend;
	int depth;
    size_t req_len;
    gboolean emit_br;
    gboolean emit_start_br = FALSE; // P.Dostal: add opening newline tag in the cycle just when closing tag will be added
    gboolean flush_empty = (flush && 0 == inlen && ! html->first_call); // P.Dostal: when flush is called on empty input, do not add new line to html

    size_t nl_start_len;
    size_t nl_end_len;
    const gchar *nl_start = NULL;
    const gchar *nl_end = NULL;
    char *last_endl = NULL;

    gint i;

    if (html->flags & GMIME_FILTER_HTML_CONVERT_NL) {
        nl_start = get_nl(html->nl_style, TRUE);
        nl_start_len = nl_start ? strlen(nl_start) + 1 : 0;
        nl_end = get_nl(html->nl_style, FALSE);
        nl_end_len = nl_end ? strlen(nl_end) + 1 : 0;
    }

	
    req_len = inlen * 2 + 6;
    if (html->flags & GMIME_FILTER_HTML_UTF8)
        req_len += PREAMBLE_SIZE + POSTAMBLE_SIZE;


	g_mime_filter_set_size (filter, req_len, FALSE);
	
	inptr = in;
	realinend = inend = in + inlen;

    /* Petr Silar: Moving back on the whole utf-8 character */
    if (inlen && (unsigned char)*(inend - 1) >= 0x80) {
        char *last = inend - 1;
        for (i = 0; i < 7 && inptr <= last; ++i, --last) {
            if ((unsigned char)*last < 0x80) {
                inend = last + 1;
                break;
            }
            if ((unsigned char)*last >= 0xc0) {
                inend = last;
                break;
            }
        }
    }

	outptr = filter->outbuf;
	outend = filter->outbuf + filter->outsize;

    if ((html->flags & GMIME_FILTER_HTML_UTF8) && html->first_call) {
        outptr = g_stpcpy (outptr, preamble);
        html->first_call = FALSE;
    }

	
	if (html->flags & GMIME_FILTER_HTML_PRE && !html->pre_open) {
		outptr = g_stpcpy (outptr, "<pre>");
		html->pre_open = TRUE;
	}

    if (NULL == last_endl && nl_start && ! flush_empty) {
        outptr = g_stpcpy (outptr, nl_start);
    }

    if (NULL == last_endl)
        last_endl = outptr;

	
	start = inptr;
	while (inptr < inend && *inptr != '\n')
		inptr++;
	
    if (inptr == inend && (start != inend || flush_empty)) {
        emit_br = FALSE;
    } else
        emit_br = TRUE;

    // There is 'start' instead of 'inptr' to allow conversion of texts not ended by eol.
	while (start < inend) {

        if (emit_start_br) {
            outptr = g_stpcpy (outptr, nl_start);
            last_endl = outptr;
            emit_start_br = FALSE;
        }


		html->column = 0;
		depth = 0;
		
		if (html->flags & GMIME_FILTER_HTML_MARK_CITATION) {
			if ((depth = citation_depth (start, inend)) > 0) {
				char font[25];
				
				/* FIXME: we could easily support multiple colour depths here */
				
				g_snprintf (font, 25, "<font color=\"#%06x\">", html->colour);
				
				outptr = check_size (filter, outptr, &outend, 25);
				outptr = g_stpcpy (outptr, font);
			} else if (*start == '>') {
				/* >From line */
				start++;
			}
		} else if (html->flags & GMIME_FILTER_HTML_CITE) {
			outptr = check_size (filter, outptr, &outend, 6);
			outptr = g_stpcpy (outptr, "&gt; ");
			html->column += 2;
		}
		
#define CONVERT_URLS_OR_ADDRESSES (GMIME_FILTER_HTML_CONVERT_URLS | GMIME_FILTER_HTML_CONVERT_ADDRESSES)
		if (html->flags & CONVERT_URLS_OR_ADDRESSES) {
			size_t matchlen, buflen, len;
			urlmatch_t match;
			
			len = inptr - start;
			
			do {
				if (url_scanner_scan (html->scanner, start, len, &match)) {
					/* write out anything before the first regex match */
					outptr = writeln (filter, start, start + match.um_so,
							  outptr, &outend);
					
					start += match.um_so;
					len -= match.um_so;
					
					matchlen = match.um_eo - match.um_so;
					
					buflen = 20 + strlen (match.prefix) + matchlen + matchlen;
					outptr = check_size (filter, outptr, &outend, buflen);
					
					/* write out the href tag */
					outptr = g_stpcpy (outptr, "<a href=\"");
					outptr = g_stpcpy (outptr, match.prefix);
					memcpy (outptr, start, matchlen);
					outptr += matchlen;
					outptr = g_stpcpy (outptr, "\">");
					
					/* now write the matched string */
                    // we need to translate the contents of the URL since it may contain '&'
                    outptr = writeln(filter, start, start + matchlen, outptr, &outend);
					html->column += matchlen;
					start += matchlen;
					len -= matchlen;
					
					/* close the href tag */
					outptr = g_stpcpy (outptr, "</a>");
				} else {
					/* nothing matched so write out the remainder of this line buffer */
					outptr = writeln (filter, start, start + len, outptr, &outend);
					break;
				}
			} while (len > 0);
		} else {
			outptr = writeln (filter, start, inptr, outptr, &outend);
		}
		
		if ((html->flags & GMIME_FILTER_HTML_MARK_CITATION) && depth > 0) {
			outptr = check_size (filter, outptr, &outend, 8);
			outptr = g_stpcpy (outptr, "</font>");
		}
		
		if (emit_br && html->flags & GMIME_FILTER_HTML_CONVERT_NL) {
			outptr = check_size (filter, outptr, &outend, nl_start_len + nl_end_len);
            if (nl_end) {
                if (last_endl == outptr && html->nl_style != G_MIME_FILTER_HTML_NL_STYLE_BR) {
                    outptr = check_size (filter, outptr, &outend, nl_start_len + nl_end_len + G_N_ELEMENTS(br));
                    outptr = g_stpcpy (outptr, br);
                }
			    outptr = g_stpcpy (outptr, nl_end);
            }
            if (nl_start) {
                emit_start_br = TRUE;
            }
		}
		
        if (emit_br) {
		    *outptr++ = '\n';
            last_endl = outptr;
        }
		
        if (inptr < inend) {
            ++inptr;
        }
		start = inptr;
		while (inptr < inend && *inptr != '\n')
			inptr++;

        emit_br = TRUE;
	}
	
	if (flush) {
		/* flush the rest of our input buffer */
		if (start < inend)
			outptr = writeln (filter, start, inend, outptr, &outend);
		
        if (html->flags & GMIME_FILTER_HTML_UTF8) {
            if (nl_end && ! flush_empty) {
                if (last_endl == outptr && html->nl_style != G_MIME_FILTER_HTML_NL_STYLE_BR) {
                    outptr = check_size (filter, outptr, &outend, nl_end_len + G_N_ELEMENTS(br));
                    outptr = g_stpcpy (outptr, br);
                    outptr = g_stpcpy (outptr, nl_end);
                } else {
                    outptr = check_size (filter, outptr, &outend, nl_end_len);
                    outptr = g_stpcpy (outptr, nl_end);
                }
            }
            outptr = check_size (filter, outptr, &outend, POSTAMBLE_SIZE + 1);
            outptr = g_stpcpy (outptr, postamble);
        }

        if (html->pre_open) {
			/* close the pre-tag */
			outptr = check_size (filter, outptr, &outend, 10);
			outptr = g_stpcpy (outptr, "</pre>");
		}


	} else if (start < realinend) {
		/* backup */
		g_mime_filter_backup (filter, start, (unsigned) (realinend - start));
	}
	
	*out = filter->outbuf;
	*outlen = outptr - filter->outbuf;
	*outprespace = filter->outpre;
}

static void
filter_filter (GMimeFilter *filter, char *in, size_t len, size_t prespace,
	       char **out, size_t *outlen, size_t *outprespace)
{
	html_convert (filter, in, len, prespace, out, outlen, outprespace, FALSE);
}

static void 
filter_complete (GMimeFilter *filter, char *in, size_t len, size_t prespace,
		 char **out, size_t *outlen, size_t *outprespace)
{
	html_convert (filter, in, len, prespace, out, outlen, outprespace, TRUE);
}

static void
filter_reset (GMimeFilter *filter)
{
	GMimeFilterHTML *html = (GMimeFilterHTML *) filter;
	
	html->column = 0;
	html->pre_open = FALSE;
    html->first_call = TRUE;
}

static const gchar *get_nl(gint32 nl_style, gboolean start)
{
    const gchar *result;
    switch (nl_style) {
        case G_MIME_FILTER_HTML_NL_STYLE_BR:
            result = start ? NULL : br;
            break;

        case G_MIME_FILTER_HTML_NL_STYLE_DIV:
            result = start ? div_start : div_end;
            break;

        case G_MIME_FILTER_HTML_NL_STYLE_P:
            result = start ? p_start : p_end;
            break;

        default:
            result = NULL;
            break;
    }

    return result;
}


/**
 * g_mime_filter_html_new:
 * @flags: html flags
 * @colour: citation colour
 *
 * Creates a new GMimeFilterHTML filter which can be used to convert a
 * plain UTF-8 text stream into an html stream.
 *
 * Returns: a new html filter.
 **/
GMimeFilter *
g_mime_filter_html_new (guint32 flags, guint32 colour, guint32 nl_style)
{
	GMimeFilterHTML *new;
	guint i;
	
	new = g_object_newv (GMIME_TYPE_FILTER_HTML, 0, NULL);
	new->flags = flags;
	new->colour = colour;
    new->nl_style = nl_style;
	
	for (i = 0; i < NUM_URL_PATTERNS; i++) {
		if (patterns[i].mask & flags)
			url_scanner_add (new->scanner, &patterns[i].pattern);
	}
	
	return (GMimeFilter *) new;
}
